import sys
import xmlrpc.client
import re
import copy
from typing import Union
from loguru import logger as loguru_logger
from .recordset import RecordSet
from .common import frozendict


class Environment(dict):
    # Todo: Add a cache system

    def __init__(
            self,
            url: str,
            username: str,
            password: str,
            db: str = None,
            auto_auth: bool = True,
            logger: loguru_logger = None,
            log_level: str = None,
            **kw
    ):
        super().__init__(**kw)

        self._url = url[:-1] if url[-1] == "/" else url
        self._username = username
        self._password = password
        self._db = db or self._extract_db_from_url()

        self.common = xmlrpc.client.ServerProxy(f"{self._url}/xmlrpc/2/common", allow_none=True)
        self.models = None

        self.requests = list()
        self.cache = {}
        self._context = frozendict()

        # Todo: make it an object
        self.user = None

        self.logger = logger if isinstance(logger, type(loguru_logger)) else loguru_logger
        if logger is None:
            self.logger.remove()
            self.logger.add(sys.stderr, level=log_level or "INFO")

        if auto_auth:
            self.authenticate()

    def __missing__(self, key):
        res = RecordSet(key, self, context=self._context)
        if self._has_missing_cache(key):
            self.cache.update({
                key: {
                    'fields': res.fields_get(),
                    'records': dict(),
                }
            })
        return res

    def __getitem__(self, key):
        assert self.user or key == 'res.users', "You must be authenticated to access models"
        return super(Environment, self).__getitem__(key)

    def __str__(self):
        return f"{self.__class__.__name__}<{id(self)}>"



    @property
    def uid(self):
        return self.user.get('id')

    @property
    def env(self):
        return self

    @property
    def context(self):
        return self._context

    @property
    def requests_count(self):
        return len(self.requests)


    # --------------------------------------------
    #                   PUBLIC
    # --------------------------------------------


    def authenticate(self):
        uid = self.common.authenticate(self._db, self._username, self._password, {})
        if uid:
            self.user = {'id': uid}
            self.models = xmlrpc.client.ServerProxy(f"{self._url}/xmlrpc/2/object", allow_none=True)
            self.user |= self['res.users'].browse(uid).read(['name', 'login'])[0]
            self.logger.info(f"Login successful on {self._url} ({self._db}) with res.users({uid}) - {self.user.get('name')} ({self.user.get('login')})")
        else:
            self.logger.error(f"Login failed on {self._url} ({self._db}) for {self._username}")

    def ref(self, xmlid):
        # xmlid = [xmlid] if isinstance(xmlid, str) else xmlid
        # Todo: add a multi version
        module, xml_id = xmlid.split(".")
        res = self['ir.model.data'].check_object_reference(module, xml_id)
        if res:
            return self[res[0]].browse(res[1])
        return None

    def with_context(self, **kw):
        self._context = self._context.copy(**kw)
        return self


    def log_request(self, recordset, *args, **kwargs):
        # Todo: add more infos about performance.
        self.logger.trace(f"Executing {args[0]} on {recordset} with args: {args[1:]} / kwargs: {kwargs}")
        self.requests.append({
            'model': recordset._name,
            'method': args[0],
            'args': args[1:],
            'kwargs': kwargs
        })


    # --------------------------------------------
    #                   PRIVATE
    # --------------------------------------------


    def _extract_db_from_url(self, url: str = None) -> str:
        url = url or self._url
        db_re = r"(https?:\/\/)?([w]{3}\.)?([\w-]*)(.\w*)([\/\w]*)"
        return re.search(db_re, url).group(3)



    # --------------------------------------------
    #                   CACHE
    # --------------------------------------------


    def _has_missing_cache(self, key):
        return key not in self.cache

    def _update_cache_records(self, key, vals_dict: dict):
        self.cache[key]['records'].update(vals_dict)

    def update_cache_records(self, key: str, vals_list: list):
        """ Transform a read or search_read result into a usable dict and update the cache """
        vals_dict = dict()
        for r in copy.deepcopy(vals_list):
            i = r.pop('id')
            vals_dict[i] = r
        self._update_cache_records(key, vals_dict)

    def delete_cached_records(self, key: str, ids: Union[int, list[int]]):
        ids = ids if isinstance(ids, list) else [ids]
        for i in ids:
            self.cache[key]['records'].pop(i)

    # def get_cached_value(self, key: str, ids: Union[int, list] = None):
    #     ids = ids if isinstance(ids, list) else [ids]
    #     return self.cache[key]['records']



import xmlrpc.client
from .common import assert_same_model, cache, log_request, model, frozendict
from typing import Union


class RecordSet:
    def __init__(self, name, env, ids: list[int] = None, context: frozendict = None):
        self._name = name
        self._env = env
        self._curr = -1
        self._ids = self._sanitize_ids(ids or list())

        """
        Each recordset has its own context, but inherits it:
            1) from the recordset it was created
            2) from the general environment 
        """
        self._context = context.copy() if context is not None else frozendict()

        self.logger = self._env.logger

    def __str__(self):
        return f"{self._name}({', '.join(map(str, self._ids))})"

    def __bool__(self):
        return bool(self._ids)

    def __len__(self):
        return len(self._ids)

    def __iter__(self):
        self._curr = -1
        return self

    def __next__(self):
        if self._curr < len(self) - 1:
            self._curr += 1
            return self._recordset(self._ids[self._curr])
        raise StopIteration

    def __getitem__(self, item):
        if isinstance(item, slice):
            return self._recordset(self._ids[item])
        elif isinstance(item, str):
            return getattr(self, item)

    def __getattr__(self, attr):
        if attr in self._env.cache.get(self._name, dict()).get('fields', list()):
            # Todo: code this instead of having a weird patch
            # self.ensure_one()
            return self._env.cache[self._name]['records'].get(self.id, dict()).get(attr, False)
        else:
            def wrapper(*args, **kw):
                return self._execute(attr, *args, **kw)
            return wrapper

    # --- Operators ---
    @assert_same_model('union')
    def __or__(self, other):
        return self._recordset(self.ids + other.ids)

    @assert_same_model('union')
    def __ior__(self, other):
        self._ids += other.ids
        return self

    def ensure_one(self):
        if len(self) != 1:
            raise ValueError(f"Expected singleton: {self}")

    # --- Properties ---
    @property
    def env(self):
        return self._env

    @property
    def ids(self):
        return self._ids

    @property
    def context(self):
        return self._context

    @property
    def id(self):
        self.ensure_one()
        return False if not self._ids else self._ids[0]

    @property
    def _model(self):
        return self._env[self._name]
        # return self._recordset(list())


    def _recordset(self, ids: Union[list[int], int]):
        ids = [ids] if isinstance(ids, int) else ids
        return self.__class__(self._name, self._env, ids, context=self._context)

    @staticmethod
    def _sanitize_ids(ids: Union[list[int], int]):
        ids = ids if isinstance(ids, list) else [ids]
        seen = set()
        return [i for i in ids if not (i in seen or seen.add(i))]

    @staticmethod
    def _format_domain(domain: list[tuple]) -> list[list]:
        return list(map(lambda e: e if isinstance(e, str) else list(e), domain))

    @log_request
    def _execute(self, method, *args, **kw):
        """ Add ids in args if method is not @model decorated"""
        if not (m := getattr(self, method, None)) or (getattr(m, '_api', None) != 'model'):
            args = [self._ids] + list(args)
        kw['context'] = self.context | kw.get('context', dict())

        try:
            return self._env.models.execute_kw(
                self._env._db,
                self._env.uid,
                self._env._password,
                self._name,
                method,
                args,
                kw,
            )

        except Exception as e:
            if isinstance(e, xmlrpc.client.Fault) and 'cannot marshal' in str(e):
                return None
            self.logger.error(f"Error while executing {self._name}.{method}():")
            self.logger.debug(f"args / kwargs:\n {args} \n {kw}")
            self.logger.error("Odoo API Response:\n" + str(e).replace('\\n', '\n'))

    # --------------------------------------------
    #                   ORM
    # --------------------------------------------


    def browse(self, ids: Union[int, list[int]]) -> "RecordSet":
        return self._recordset(ids)

    def fields_get(self, attributes: list[str] = None):
        attributes = attributes or ['string', 'help', 'type']
        return self._execute('fields_get', attributes=attributes)

    @model
    def check_object_reference(self, module, xml_id):
        # Extra safety check, don't delete (even if nobody should use it directly)
        if self._name != 'ir.model.data':
            return self['ir.model.data'].check_object_reference(module, xml_id)
        return self._execute('check_object_reference', module, xml_id)

    def with_context(self, **kw):
        self._context = self._context.copy(**kw)
        return self


    # --- CRUD ---


    @model
    def search(self, domain: list[tuple], **kw):
        ids = self._execute('search', self._format_domain(domain), **kw)
        return self._recordset(ids)

    def read(self, fields: list[str] = None, **kw):
        fields = fields or list()
        res = self._execute('read', fields=fields, **kw)
        self._env.update_cache_records(self._name, res)
        return res

    @model
    def search_read(self, domain: list[tuple], fields: list[str] = None, **kw):
        fields = fields or list()
        res = self._execute('search_read', self._format_domain(domain), fields=fields, **kw)
        self._env.update_cache_records(self._name, res)
        return self._recordset([r.get('id') for r in res])

    @model
    def search_count(self, domain: list[tuple]):
        return self._execute('search_count', self._format_domain(domain))

    @model
    def create(self, vals_list: Union[dict, list[dict]]):
        vals_list = vals_list if isinstance(vals_list, list) else [vals_list]
        ids = self._execute('create', vals_list)
        return self._recordset(ids)

    def write(self, vals: dict):
        return self._execute('write', vals)

    @cache('delete')
    def unlink(self):
        res = self._execute('unlink')
        if res:
            self._env.delete_cached_records(self._name, self._ids)
        return res

    def copy(self):
        res_id = self._execute('copy')
        return self._recordset(res_id)


    # --- ORM helpers ---


    def mapped(self, field: str):
        return [getattr(rec, field) for rec in self]

    def filtered(self, func: Union[callable, str]):
        if isinstance(func, str):
            name = func
            func = lambda rec: rec[name]
            # func = lambda rec: any(rec.mapped(name))      # Odoo behavior
        return self.browse([rec.id for rec in self if func(rec)])

    def filtered_domain(self, domain: list[tuple]):
        return self.search([('id', 'in', self.ids)] + domain)




# --------------------------------------------
#                DECORATORS
# --------------------------------------------

def model(fn):
    fn._api = 'model'
    return fn

def log_request(fn):
    def wrapper(self, *args, **kwargs):
        # Todo: add more infos about performance.
        self.env.log_request(self, *args, **kwargs)
        return fn(self, *args, **kwargs)
    return wrapper

def assert_same_model(op=None):
    def decorator(fn):
        def wrapper(self, other):
            if self._name != other._name:
                raise ValueError(f"Mixed apples and oranges => cannot {op or fn.__name__} on {self} and {other}")
            return fn(self, other)
        return wrapper
    return decorator

def cache(op=None):
    def decorator(fn):
        def wrapper(self, *args, **kwargs):
            print(f"Cache {op or fn.__name__} on {self}")
            return fn(self, *args, **kwargs)
        return wrapper
    return decorator



# --------------------------------------------
#               frozendict stuff
# --------------------------------------------
def remove(fn):
    def decorator(self, *args, **kwargs):
        raise NotImplementedError(f'{fn.__name__} not allowed on {self.__name__}')
    return decorator


class frozendict(dict):
    """ A class that allow you to lock dictionary so no option will be available other than get """

    @remove
    def __delitem__(self, *args) -> None:
        ...

    @remove
    def __setitem__(self, *args) -> None:
        ...

    @remove
    def clear(self, *args) -> None:
        ...

    @remove
    def pop(self, *args) -> None:
        ...

    @remove
    def popitem(self, *args) -> None:
        ...

    @remove
    def setdefault(self, *args) -> None:
        ...

    @remove
    def update(self, *args) -> None:
        ...

    def copy(self, **kw) -> "frozendict":
        return frozendict({
            **self,
            **kw,
        })




from setuptools import find_packages, setup

with open("README.md", "r") as f:
    long_description = f.read()

setup(
    name="otools-rpc",
    version="0.3.0",
    description="A tool to interact with Odoo's external API.",
    packages=find_packages(exclude=["tests"]),
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/MrFaBemol/otools-rpc",
    author="Gautier Casabona",
    author_email="gcasabona.pro@gmail.com",
    license="MIT",
    classifiers=[
        "Development Status :: 3 - Alpha",
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python :: 3.8",
        "Operating System :: OS Independent",
        "Intended Audience :: Developers",
        "Natural Language :: English"
    ],
    install_requires=[
        "loguru >= 0.7.0",
        "requests == 2.29.0"  # Higher version can break your docker python lib
    ],
    extras_require={
        "dev": ["pytest>=7.0", "twine>=4.0.2"],
    },
    python_requires=">=3.8",
)



from loguru import logger
import requests
import xmlrpc.client


class DBManager:
    
    def __init__(self, url: str, password: str):

        # --------------------------------------------
        #                   PRIVATE
        # --------------------------------------------
        
        self._url = url[:-1] if url[-1] == "/" else url
        self._password = password

        self._duplicate_url = '{}/web/database/duplicate'.format(self._url)
        self._drop_url = '{}/web/database/drop'.format(self._url)
        
        
        # --------------------------------------------
        #                   PUBLIC
        # --------------------------------------------
        
        self.dbobject = xmlrpc.client.ServerProxy('{}/xmlrpc/2/db'.format(self._url))
        
    def duplicate(self, db, new_name):
        res = requests.post(self._duplicate_url, data={'master_pwd': self._password, 'name': db, 'new_name': new_name})
        return res
    
    def drop(self, db):
        res = requests.post(self._drop_url, data={'master_pwd': self._password, 'name': db})
        return res

    def list(self):
        return self.dbobject.list()
